from typing import List, Union

import numpy as np

from payton.math.vector import Vector3D

DIFF = 0.0000001

Matrix = Union[List[Vector3D], np.ndarray]

IDENTITY_MATRIX: Matrix = [
    [1.0, 0.0, 0.0, 0.0],
    [0.0, 1.0, 0.0, 0.0],
    [0.0, 0.0, 1.0, 0.0],
    [0.0, 0.0, 0.0, 1.0],
]

NP_IDENTITY_MATRIX = np.array(IDENTITY_MATRIX, dtype=np.float32)


def bullet_to_matrix(quaternion: List[float]) -> Matrix:
    """Turn Bullet Quaternion to matrix.

    Transforms3d
    ============

    The transforms3d package, including all examples, code snippets and attached
    documentation is covered by the 2-clause BSD license.

    Copyright (c) 2009-2019, Matthew Brett and Christoph Gohlke
    All rights reserved.
    """
    w, x, y, z = quaternion
    Nq = w * w + x * x + y * y + z * z
    if Nq < DIFF:
        return IDENTITY_MATRIX
    s = 2.0 / Nq
    X = x * s
    Y = y * s
    Z = z * s
    wX = w * X
    wY = w * Y
    wZ = w * Z
    xX = x * X
    xY = x * Y
    xZ = x * Z
    yY = y * Y
    yZ = y * Z
    zZ = z * Z
    return [
        [1.0 - (yY + zZ), xY - wZ, xZ + wY, 0],
        [xY + wZ, 1.0 - (xX + zZ), yZ - wX, 0],
        [xZ - wY, yZ + wX, 1.0 - (xX + yY), 0],
        [0, 0, 0, 1],
    ]



from typing import Any, Dict, List, Optional

import numpy as np

from payton.math.functions import cubemap_projection_matrices
from payton.math.vector import Vector3D


class Light:
    def __init__(
        self,
        position: Optional[Vector3D] = None,
        color: Optional[Vector3D] = None,
        **kwargs: Any,
    ):
        """Initialize light

        Light in Payton is a point light to all directions

        Keyword arguments:
        position -- Position of the light in space
        color -- Color of the light source
        """
        self._position = [10.0, 7.0, 6.0] if position is None else position
        self._color = [1.0, 1.0, 1.0] if color is None else color
        self._position_np: np.ndarray = np.array(list(self._position), dtype=np.float32)
        self._color_np: np.ndarray = np.array(list(self._color), dtype=np.float32)
        self._shadow_matrices: List[np.ndarray] = []
        self._shadow_far_plane = 100.0

        self.active: bool = True

    @property
    def position(self) -> Vector3D:
        """Return the position of the light"""
        return self._position

    @position.setter
    def position(self, position: Vector3D) -> None:
        """Set the position of the light

        Keyword arguments:
        position -- Position in space
        """
        self._position = position
        self._position_np = np.array(self.position, dtype=np.float32)
        self._shadow_matrices = []

    def to_dict(self) -> Dict[str, Any]:
        """Convert the light into dictionary"""
        return {
            "position": self.position,
            "color": self.color,
            "active": self.active,
        }

    @property
    def shadow_matrices(self) -> List[np.ndarray]:
        """Return the shadow casting cubemap projection matrices for the light"""
        if len(self._shadow_matrices) > 0:
            return self._shadow_matrices
        self._shadow_matrices = cubemap_projection_matrices(self.position, self._shadow_far_plane)
        return self._shadow_matrices

    @property
    def shadow_far_plane(self) -> float:
        """Return the furthest distance that the light can cast shadows"""
        return self._shadow_far_plane

    @shadow_far_plane.setter
    def shadow_far_plane(self, distance: float) -> None:
        """Set the furthest distance that the light can cast shadows

        Keyword arguments:
        val -- Distance
        """
        self._shadow_far_plane = distance
        self._shadow_matrices = []

    @property
    def color(self) -> Vector3D:
        """Return the light color"""
        return self._color

    @color.setter
    def color(self, color: Vector3D) -> None:
        """Set the light color

        Keyword arguments:
        color -- Color of the light
        """
        self._color = color
        self._color_np = np.array(self._color, dtype=np.float32)



# pylama:ignore=W
"""#Payton GUI Module
"""
from payton.scene.gui.base import Hud, Rectangle, Shape2D, Text
from payton.scene.gui.help import info_box
from payton.scene.gui.window import Button, EditBox, Panel, Theme, Window, WindowAlignment

__all__ = [
    "Button",
    "EditBox",
    "Hud",
    "info_box",
    "Panel",
    "Rectangle",
    "Shape2D",
    "Text",
    "Theme",
    "Window",
    "WindowAlignment",
]



from typing import List, Optional, Tuple, Union

import numpy as np
import pyrr

from payton.math.functions import distance as distance_native
from payton.math.types import GArray
from payton.math.vector import Vector3D

DIFF = 0.0000001

"""Few notes about this module.

Most of the methods defined in this module expects and returns Numpy Arrays.
Numpy Arrays are simply C Type Arrays, where all the bytes are packed and ordered
properly. Regular Python arrays have some extra bytes that doesn't play well with
the expectations of Graphics card.

In order to give the graphics card the data type it expects, Numpy Arrays are
ideal"""


def point_project(p: GArray, origin: GArray, direction: GArray) -> float:
    return direction[0] * (p[0] - origin[0]) + direction[1] * (p[1] - origin[1]) + direction[2] * (p[2] - origin[2])


def distance(v1: np.ndarray, v2: np.ndarray) -> float:
    """Distance between two vectors defined as Numpy arrays

    Keyword arguments:
    v1 -- First vector
    v2 -- Second vector
    """
    v3 = v2 - v1
    return pyrr.vector3.length(v3)


def distance2(v1: np.ndarray, v2: np.ndarray) -> float:
    """Square of the distance between two given vectors defined as Numpy arrays

    Keyword arguments:
    v1 -- First vector
    v2 -- Second vector"""
    v1 = v1[:3]
    v2 = v2[:3]
    v3 = v2 - v1
    return pyrr.vector3.length(v3) ** 2


def combine(v1: GArray, v2: GArray, f1: float, f2: float) -> np.ndarray:
    """Combines two vectors given by f1 and f2 ratios and returns Numpy array

    Keyword arguments:
    v1 -- First vector
    v2 -- Second vector
    f1 -- First vector's ratio
    f2 -- Second vector's ratio
    """
    x = (f1 * v1[0]) + (f2 * v2[0])
    y = (f1 * v1[1]) + (f2 * v2[1])
    z = (f1 * v1[2]) + (f2 * v2[2])

    return np.array([x, y, z], dtype=np.float32)


def combine3(v1: GArray, v2: GArray, v3: GArray, f1: float, f2: float, f3: float) -> np.ndarray:
    """Combine three vectors with three factors

    Keyword arguments:
    v1 -- First vector
    v2 -- Second vector
    v3 -- Third vector"""
    res = [
        (f1 * v1[0]) + (f2 * v2[0]) + (f3 * v3[0]),
        (f1 * v1[1]) + (f2 * v2[1]) + (f3 * v3[1]),
        (f1 * v1[2]) + (f2 * v2[2]) + (f3 * v3[2]),
        0,
    ]
    return np.array(res, dtype=np.float32)


def raycast_sphere_intersect(start: GArray, direction: GArray, sphere_center: np.ndarray, sphere_radius: float) -> bool:
    """Check if ray intersects given sphere, returns boolean

    Keyword arguments:
    start -- Start of the ray vector
    direction -- Direction vector of the given vector
    sphere_center -- Center of the sphere to test
    sphere_radius -- Radius of the sphere to test"""
    proj = point_project(sphere_center, start, direction)
    if proj < 0:
        proj = 0.0
    vc = combine(start, direction, 1.0, proj)
    dist = distance2(sphere_center, vc)
    return dist < (sphere_radius ** 2)


def _find_box_mid_for_intersection(start: GArray, box_a: GArray, box_b: GArray) -> Tuple[List[bool], List[float], bool]:
    is_middle = [False, False, False]
    plane = [0.0, 0.0, 0.0]
    result = True
    for i in range(3):
        if start[i] < box_a[i]:
            plane[i] = box_a[i]
            is_middle[i] = False
            result = False
        elif start[i] > box_b[i]:
            plane[i] = box_b[i]
            is_middle[i] = False
            result = False
        else:
            is_middle[i] = True
    return is_middle, plane, result


def raycast_box_intersect(start: GArray, direction: GArray, box_a: GArray, box_b: GArray) -> Optional[Vector3D]:
    """Check if ray intersects given axis-aligned-box, returns point of intersection if True

    AABB - Axis Aligned Bounding Box means the given box is parallel to the X,Y,Z axises.

    Keyword arguments:
    start -- Start of the ray vector
    direction -- Direction vector of the given vector
    box_a -- Starting vector of the box
    box_b -- Ending vector of the box"""
    max_dist = [0.0, 0.0, 0.0]

    is_middle, plane, result = _find_box_mid_for_intersection(start, box_a, box_b)

    plane_id = 0
    for i in range(3):
        if is_middle[i] or direction[i] == 0.0:
            max_dist[i] = -1
        else:
            max_dist[i] = (plane[i] - start[i]) / direction[i]
            if max_dist[i] > 0.0:
                if max_dist[plane_id] < max_dist[i]:
                    plane_id = i
                result = True

    if result:
        tmp = [0.0, 0.0, 0.0]
        for i in range(3):
            if plane_id == i:
                tmp[i] = plane[i]
            else:
                tmp[i] = start[i] + max_dist[plane_id] * direction[i]
                result = (tmp[i] >= box_a[i]) and (tmp[i] <= box_b[i])
                if not result:
                    return None

        return tmp
    return None


def point_on_line(point: Vector3D, start: Vector3D, end: Vector3D) -> bool:
    """Test if the given point is on the line defined by start-end vectors

    Keyword arguments:
    point -- Point to test
    start -- Line start
    end -- Line end
    """
    ab = distance_native(end, start)
    ap = distance_native(point, start)
    bp = distance_native(end, point)
    return abs(ap + bp - ab) < DIFF


def raycast_plane_intersect(
    start: GArray, direction: GArray, plane_point: GArray, plane_normal: GArray
) -> Optional[np.ndarray]:
    """Test if the given ray intersects with the defined plane, returns
    intersection point if True else None

    Keyword arguments:
    start -- Start of the ray vector
    direction -- Direction of the ray vector
    plane_point -- A point on the plane surface to locate the plane
    plane_normal -- Normal of the defined plane"""
    start = start[:3]
    direction = direction[:3]
    plane_point = plane_point[:3]
    plane_normal = plane_normal[:3]
    d = np.dot(direction, plane_normal)  # type: ignore
    res = (d > DIFF) or (d < -DIFF)
    if not res:
        return None
    sp = np.subtract(plane_point, start)
    d = 1.0 / d
    t = np.dot(sp, plane_normal) * d  # type: ignore
    if t <= 0:
        return None
    return combine(start, direction, 1.0, t)


def raycast_triangle_intersect(
    start: GArray, direction: GArray, p1: GArray, p2: GArray, p3: GArray
) -> Tuple[Union[np.ndarray, None], Union[np.ndarray, None]]:
    """Returns intersection point, intersection normal

    Keyword arguments:
    start -- Start of the ray vector
    direction -- Direction of the ray vector
    p1 -- First corner coordinates of the triangle
    p2 -- Second corner coordinates of the triangle
    p3 -- Third corner coordinates of the triangle
    """
    v1 = np.subtract(p2, p1)
    v2 = np.subtract(p3, p1)
    pvec = np.cross(direction, v2)
    det = np.dot(v1, pvec)  # type: ignore
    if (det < DIFF) and (det > -DIFF):
        return None, None
    inv_det = 1.0 / det
    tvec = np.subtract(start, p1)
    u = np.dot(tvec, pvec) * inv_det  # type: ignore
    if u < 0 or u > 1:
        return None, None
    qvec = np.cross(tvec, v1)
    v = np.dot(direction, qvec) * inv_det  # type: ignore
    res = (v > 0) and (u + v <= 1.0)
    if not res:
        return None, None
    t = np.dot(v2, qvec) * inv_det  # type: ignore
    if t <= 0:
        return None, None
    ip = combine(start, direction, 1, t)
    inor = np.cross(v1, v2)
    inor /= np.linalg.norm(inor)  # type: ignore
    return ip, inor


def line_triangle_intersect(
    start: np.ndarray,
    end: np.ndarray,
    p1: np.ndarray,
    p2: np.ndarray,
    p3: np.ndarray,
) -> bool:
    """Check if line intersects triangle, return boolean

    NOTE: Line does not mean Ray!! Ray goes to infinite. Line is finite.

    Keyword arguments:
    start -- Start of the given line segment
    end -- End of the given line segment
    p1 -- First corner coordinates of the triangle
    p2 -- Second corner coordinates of the triangle
    p3 -- Third corner coordinates of the triangle"""
    direction = end - start
    norm = np.linalg.norm(direction)  # type: ignore
    if norm == 0:
        return False
    direction /= norm
    ip, _inor = raycast_triangle_intersect(start, direction, p1, p2, p3)
    if ip is None:
        return False
    # a---c----b
    d1 = distance(start, ip)
    d2 = distance(ip, end)
    d3 = distance(start, end)
    diff = d3 - (d1 + d2)
    return diff < DIFF and diff > -DIFF



from PIL import Image, ImageDraw

from payton.scene.gui.base import Hud
from payton.scene.gui.window import Button, Window, WindowAlignment


def help_win() -> Window:
    text = """Payton Shortcuts and Mouse Controls:

MOUSE:
Zoom In-Out: Mouse Wheel
Rotate: Right Click and Drag
Pan: Middle Click and Drag
KEYBOARD:
ESC: Quit
H: Show/Hide Help
C: Change Camera mode (Perspective/Orthographic)
Space: Start/Pause Scene Animations/Clocks
G: Show/Hide Grid
W: Display mode (wireframe/solid)
F2: Previous Camera F3: Next Camera
    """
    help_window = Window(
        align=WindowAlignment.RIGHT,
        title="How to use Payton Scene",
        height=10000,
        width=520,
    )
    help_window.add_child("text", Button(width=500, height=400, left=10, top=40, label=text))
    return help_window


def info_box(left: int, top: int, label: str) -> Hud:
    """Create Info Box along with it's HUD so you do not need to setup the
    whole HUD.

    Keyword arguments:
    left -- Left position of the info box
    top -- Top position of the info box
    label -- Label / text of the info box
    """
    hud = Hud()
    timg = Image.new("RGBA", (1, 1))
    d = ImageDraw.Draw(timg)
    res = d.textsize(label, font=hud.font)
    width = res[0] + 60
    height = res[1] + 60

    window = Window("Info", width=width, height=height + 60, left=left, top=top)

    window.add_child(
        "label",
        Button(label=label, width=width - 20, height=height - 40, top=40, left=10),
    )
    window.add_child(
        "close",
        Button(
            label="CLOSE",
            left=int(width / 2) - 10,
            top=height + 10,
            width=int(width / 2),
            height=30,
            on_click=window.hide,
        ),
    )
    hud.add_child("window", window)
    return hud



"""
Animated Wavefront Pack format.

Animated Wavefront Pack 3D files (AWP3D) are simply ZIP
files with each frame packed inside.

Easiest way to generate those files is through using Blender Export
Add-On.
"""
import os
import time
from typing import Any, Dict, List, Optional, Tuple
from zipfile import ZipFile

import numpy as np

from payton.scene.geometry import Wavefront
from payton.scene.material import POINTS
from payton.scene.shader import DEFAULT_SHADER, PARTICLE_SHADER, Shader
from payton.tools.bar import progress


class AWP3D(Wavefront):
    """
    AWP3D File Format Loader.

    This file format is quite like MD2 but supports high-poly animations.
    Simply, each awp3d file is a zip file with every frame of an anination
    is exported as Wavefront and packed.

    AWP3D loads the zip file and parses each Wavefront. Then loads them as
    key frames in Payton
    """

    def __init__(self, filename: str = "", fps: int = 30, **kwargs: Any) -> None:
        """Initialize AWP3D."""
        super().__init__(**kwargs)
        self.frames: List[Wavefront] = []
        self._frame: int = 0
        self.num_frames: int = 0
        self._time: float = 0
        self._path: str = ""
        self._loop: bool = False
        self._from_frame = 0
        self._to_frame = 0
        self._active_framecount = 0
        self.fps = fps
        self._frame_period = 1.0 / fps
        self.animate = True
        self._time = 0.0
        self.animations: Dict[str, Tuple[int, int]] = {}

        if os.path.exists(filename):
            self.load_file(filename)

    def start(self) -> None:
        """Manually start animation."""
        self.animate = True

    def pause(self) -> None:
        """Pause animation."""
        self.animate = False

    def stop(self) -> None:
        """Stop animation."""
        self.animate = False
        self._time = 0.0

    def set_range(self, from_frame: int, to_frame: int) -> None:
        """Set animation frame range.

        Keyword arguments:
        from_frame -- Starting frame number
        to_frame -- Ending frame number
        """
        if from_frame >= self.num_frames:
            raise BaseException("Frame range error - from_frame out of bounds")
        if to_frame >= self.num_frames:
            raise BaseException("Frame range error - to_frame out of bounds")
        if from_frame < 0 or to_frame < 0:
            raise BaseException("Frame range error - Negative value defined")
        self._from_frame = from_frame
        self._to_frame = to_frame
        self._active_framecount = to_frame - from_frame
        self._frame_period = 1.0 / self.fps

    def set_animation(self, name: str, from_frame: int, to_frame: int) -> None:
        """Create an animation definition.

        Keyword arguments:
        name -- Name of the animation
        from_frame -- Animation starting frame number
        to_frame -- Animation ending frame number
        """
        self.animations[name] = (from_frame, to_frame)

    def run_animation(self, name: str) -> None:
        """Run the selected animation.

        Keyword arguments:
        name -- Name of the desired animation
        """
        if name not in self.animations:
            raise BaseException("Animation not defined")
        self.set_range(self.animations[name][0], self.animations[name][1])

    def load_file(self, filename: str) -> bool:
        """Load file into system.

        NOTE: This takes a while, that's why there is a loading
        indicator here. If it is annoying for you, ping me thus
        I can remove ;)

        Keyword arguments:
        filename -- File name to load
        """
        if not os.path.exists(filename):
            raise BaseException(f"File not found: {filename}")
        self._path = os.path.dirname(os.path.abspath(filename))
        archive = ZipFile(filename, 'r')
        files = archive.namelist()
        max_frame = max([int(f.split('.')[0]) for f in files]) + 1
        min_frame = min(int(f.split('.')[0]) for f in files)
        for f in range(min_frame, max_frame):
            progress(f, max_frame - 1)
            wobj = Wavefront()
            wobj.path = self._path
            data = archive.read(f"{f}.obj").decode('utf-8')
            material = archive.read(f"{f}.mtl").decode('utf-8')
            wobj.load_material(material)
            wobj.load(data)
            self.frames.append(wobj)
        self._active_framecount = max_frame - min_frame
        self.num_frames = max_frame - min_frame
        archive.close()
        return True

    def render(
        self, lit: bool, shader: Shader, parent_matrix: Optional[np.ndarray] = None, _primitive: int = None
    ) -> None:
        """Render AWP3D."""
        if not self._visible:
            return
        self.update_matrix(parent_matrix=parent_matrix)
        self.track()

        if not self.animate:
            self.frames[0].render(lit, shader, self._model_matrix)
            return

        if self._time == 0:
            self._time = time.time()

        if time.time() - self._time >= self._frame_period:
            self._frame = (self._frame + 1) % self._active_framecount
            self._time = time.time()
        render_frame = self._frame + self._from_frame
        self.frames[render_frame].render(lit, shader, self._model_matrix)

    def toggle_wireframe(self) -> None:
        """Toggle wireframe view of the Object."""
        d = (self.material.display + 1) % 3

        for mat in self.materials.values():
            mat.display = d
            mat.particle_size = 0.01

        self.shader = PARTICLE_SHADER if d == POINTS else DEFAULT_SHADER
        self.refresh()
        for frame in self.frames:
            frame.toggle_wireframe()
